import mysql.connector
from mysql.connector import errorcode, pooling
import json
from datetime import datetime
import os
# from .configs.database import connections 

import importlib

config_location = os.getenv("CONFIG_PATH", "KassOrm.configs.database")
connections = importlib.import_module(config_location).connections


class Conn:       
    
    def __init__(self, conn = None) -> None:    
        
        self.query = None          
        self.params = None           
        
        config_pool = connections['default'] if conn == None else connections[conn]        
        
        try:
            self.conn_pool = mysql.connector.pooling.MySQLConnectionPool(**config_pool)    
            
        except mysql.connector.Error as err:
            if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                print("Something is wrong with your user name or password")
            elif err.errno == errorcode.ER_BAD_DB_ERROR:
                print("Database does not exist")
            else:
                print(err)
          
              
    def getQuery(self, query, params):
        self.query = query
        self.params = params
        return self
        
        
    def execute(self, fetchall = True):
        
        return self.execute_select(fetchall)
            
    
    def execute_select(self, fetchall = True):
        
        try:
            conn = self.conn_pool.get_connection()
                
            if conn and conn.is_connected():
                cursor = conn.cursor(dictionary=True, buffered=True)
                cursor.execute(self.query, params=self.params)                
            
                                
                if cursor.with_rows: 
                    result = cursor.fetchall() if fetchall == True else cursor.fetchone()                                                                   
                    cursor.close() 
                    conn.close()  
                  
                else:
                    result = None    
                
            else:
                print('Not connected. Check the credentials or the server')
                result = None
        
        except Exception as e:
            print(f"Error to catch data: {e}")
            result = None

        return result
        
        
    def execute_insert(self, many = False):
        
        conn = self.conn_pool.get_connection()
        if conn and conn.is_connected():
            cursor = conn.cursor(dictionary=True, buffered=True)
            
            if many:
                cursor.executemany(self.query, self.params)  
                
                lastid = cursor.lastrowid
                ultimos_ids = []
                x=0
                for _ in self.params:
                    id = lastid + x
                    ultimos_ids.append(id)
                    x=+1
            else:
                cursor.execute(self.query, params=self.params)  
                ultimos_ids = cursor.lastrowid
                
            conn.commit()           
            

            cursor.close()
            conn.close() 
            
        return ultimos_ids              
        
        
    def execute_update(self):
        conn = self.conn_pool.get_connection()
        if conn and conn.is_connected():
            cursor = conn.cursor(dictionary=True, buffered=True)

            cursor.execute(self.query, self.params)             
                
            conn.commit()   

            cursor.close()
            conn.close() 
            
        return True     


    def execute_delete(self):
        conn = self.conn_pool.get_connection()
        if conn and conn.is_connected():
            cursor = conn.cursor(dictionary=True, buffered=True)

            cursor.execute(self.query, self.params)             
                
            conn.commit()   

            cursor.close()
            conn.close() 
            
        return True     


    def execute_create(self):
        conn = self.conn_pool.get_connection()
        if conn and conn.is_connected():
            cursor = conn.cursor(dictionary=True, buffered=True)

            cursor.execute(self.query, self.params)             
                
            conn.commit()   

            cursor.close()
            conn.close() 
            
        return True   


    def __test_pool(self):
        try:
            # Criar várias conexões simultâneas
            for x in range(10):
                conn = self.conn_pool.get_connection()
                if conn and conn.is_connected():
                    print("Conexão obtida do pool com sucesso. " + str(x) )
                    # Realize operações na conexão se necessário                 
                    conn.close()  # Importante: fechar a conexão para devolvê-la ao pool
                else:
                    print("Não foi possível obter a conexão do pool.")

        except Exception as e:
            print(f"Erro ao testar o pool de conexões: {e}")

        finally:      
            # Certifique-se de fechar o pool de conexões após o teste
            conn_in_pool = self.conn_pool._remove_connections()
            print("Closed: " + str(conn_in_pool))


import json
from .Conn import Conn
from datetime import datetime
from .configs.database import useSofdelete

class Querier:
    
    def __init__(self, conn=None, softDelete = False) -> None:   
        
        self.__conn = Conn(conn) 
        
        __useSoftDelete = useSofdelete['default'] if conn == None else useSofdelete[conn]     
        
        self.__softDelete = softDelete if softDelete == False else __useSoftDelete
        
        self.__withTrashed = False
        
        self.__params = {}    
        
        self.__columns = [{"name":"*"}]
        
        self.__table = None
        
        self.__table_alias = None
        
        self.__limit = None
        
        self.__offset = None
    
        self.__group = None 
               
        self.__order = None        
        
        self.__conditional = []
        
        self.__SQL = ""        
        
        self.__values_update = None
        
        self.__last_method = 'select'
        
        self.__join = []
        
        self.__queries = []
    
       
    # raw---------------------------------------------------------------
    def raw(self, query:str, params:dict = {}, all:bool = True):        
        result =  self.__conn.getQuery(query,params).execute(all)      
        data = json.dumps(result, default=self.__convert_datetime_to_string)    
        return data    
    
    
    #select-------------------------------------------------------------
    
    def select(self, columns:list[str|dict]):
        """Informar quais colunas buscar""" 
        
        self.__columns = []
    
        for col in columns:
            
            if type(col) == dict:
                self.__columns.append(col) 
            else:
                self.__columns.append({"name":col}) 
            
        return self
        
    
    def table(self, table:str, alias:str = None):
        """Informa a table e seu alias se houver"""
                
        self.__table = table
        
        self.__table_alias = alias
        
        # self.__columns = self.__getColumnsNames(self.__table, alias)
        
        return self
    
    
    def withTrashed(self):
        self.__withTrashed = True
        return self
    
    
    
    
    #joins-------------------------------------------------------------
        
    def join(self, table, model_key, local_key):
        self.__join.append(f" LEFT JOIN {table} ON ({table}.{model_key} = {self.__table}.{local_key}) ")    
        return self
    
    
    
    
    # condicional-------------------------------------------------------------
        
    def where(self, values:dict, conector:str = 'AND'):
                
        where = ""     
               
        for col, value in values.items():
            keyParamValue, keyParam = self.__paramKey(col)
            self.__params[keyParamValue] = value
            
            where += f" {col} = {keyParam} AND "
            
        where = where[:-4]    
            
        self.__conditional.append({"where":where, "conector":conector})    
        return self
            
    
    def whereIn(self, column:str, values:list[str], conector:str = "AND", type:str = ''):
        
        where = f"{column} {type} IN "
        
        vals = ''
        for index, val in enumerate(values):
            keyParamValue, keyParam = self.__paramKey(f"{column}in{index}")
            self.__params[keyParamValue] = val
        
            vals += f"'{keyParam}', "
            
        where += f"({vals[:-2]})"        
    
        self.__conditional.append({"where":where, "conector":conector})    
        return self
    
    
    def whereNotIn(self, column:str, values:list[str], conector = 'AND'):
        
        return self.whereIn(column=column, values=values, type="NOT", conector=conector)  
    
    
    def orWhere(self, values:dict):
        
        return self.where(values, "OR")
    
    
    def orWhereIn(self, column:str, values:list[str]):
        
        return self.whereIn(column, values, "OR")
            
    
    def whereIsNull(self, column, conector:str = 'AND', type:str=""):        

        where = f" {column} IS {type} NULL "    
        self.__conditional.append({"where":where, "conector":conector})    
        return self
    
    
    def whereIsNotNull(self, col, conector:str = 'AND'):
        
        return self.whereIsNull(col, conector, type="NOT")
    
    
    def whereLike(self, col:str, value:str, conector:str="AND"):
        
        keyParamValue, keyParam = self.__paramKey(f"like{col}")
        self.__params[keyParamValue] = value
        
        where = f" {col} LIKE {keyParam} " 
            
        self.__conditional.append({"where":where, "conector":conector})    
        return self
    
    
       
    # pos condicional   -------------------------------------------------------------         
    
    def limit(self, limit:int):
        self.__limit = limit
        return self
    
    
    def offset(self, offset:int):
        self.__offset = offset
        return self
    
    
    def groupBy(self, group:str|list[str]):  
        
        if type(group) == str:    
            grouped = group
            
        else:            
            grouped = ""
            for gr in group:
                grouped += f"{gr}, "
            grouped = grouped[:-2] 
            
        
        self.__group = grouped
            
        return self
    
    
    def orderBy(self, order:str|list[str]):  
    
        if type(order) == str:    
            ordered = order
            
        else:            
            ordered = ""
            for col, dir in order.items():        
                ordered += f"{col} {dir}, "
            ordered = ordered[:-2] 
            
        
        self.__order = ordered
            
        return self
     
    

    
    
        
    # insert update delete-------------------------------------------------------------
    def insert(self, data:dict|list[dict]):    
        
        self.__construct_insert_query(data)  
        
        query = self.__conn.getQuery(self.__SQL, data)
  
        type_insert = False if type(data) == dict else True
        
        return query.execute_insert(type_insert)    
        
        
    def update(self, data:dict):   
        
        self.__construct_update_query(data)        

        return self.__conn.getQuery(self.__SQL,  self.__params).execute_update()    
       

    def delete(self):        
       
        if self.__softDelete == False:
            self.__construct_delete_query()
            return self.__conn.getQuery(self.__SQL,  self.__params).execute_delete()    
        else:   
            now = datetime.now()
            self.__construct_update_query({self.__softDelete:now})
            return self.__conn.getQuery(self.__SQL,  self.__params).execute_update()   
       
       
    def active(self): 
        if self.__withTrashed:
            self.__construct_update_query({self.__softDelete:None})
            return self.__conn.getQuery(self.__SQL,  self.__params).execute_update()   

    



    
    # finalização-------------------------------------------------------------    
    
    def get(self):
        
        self.__construct_select_query() 
        
        result =  self.__conn.getQuery(self.__SQL, self.__params).execute()      
        data = json.dumps(result, default=self.__convert_datetime_to_string)    
        return json.loads(data)      
       
    
    def first(self):
        self.__construct_select_query()      
        
        result = self.__conn.getQuery(self.__SQL, self.__params).execute(False) 
        data = json.dumps(result, default=self.__convert_datetime_to_string)    
        return data       

    
    def toSql(self):
        
        if self.__last_method == 'select':              
            self.__construct_select_query()
            
        return self.__SQL
    
    
    def toInfo(self):
        if self.__last_method == 'select':     
            self.__construct_select_query()
            
        return {'conn':self.__conn, 'type':self.__last_method, 'queries':self.__queries, 'params':self.__params}
    
    
    def toParams(self):
        return self.__params
        
    
   
    
    
    
    
    
    
    # misc-------------------------------------------------------------
    
    
    def __construct_select_query(self):    
        
        sql = "SELECT "    
        
        
        #colunas
        tableColAlias = "" if self.__table_alias == None else self.__table_alias + "."
        
        columns = ""
        for col in self.__columns:
            
            colAlias = ''
            if 'alias' in col:
                if col['alias'] != '':
                    colAlias = f" AS {col['alias']}"
            
            columns += f"{tableColAlias}{col['name']}{colAlias}, "             
        columns = columns[:-2]           
        if columns != "":
            sql += columns.strip() +" "
        
        
        
        #alias       
        tableAlias = "" if self.__table_alias == None else self.__table_alias       
        if columns != "":
            sql += " FROM " + self.__table +" "
        else:
            sql += f" FROM {self.__table} {tableAlias}" +" "
        
        
      
        
        
        
        #condicional
        if self.__softDelete != False:
            if self.__withTrashed == False:
                self.whereIsNull(self.__softDelete)
          

            
        where = ""
        for wheres in self.__conditional: 
            queryWhere = wheres['where']      
            where += f" {wheres['conector']} ({queryWhere})"            
        if where != '' :
            where = f" WHERE {where} "  
        where = where.replace("WHERE  AND","WHERE").replace("WHERE  OR","WHERE")
        where = where.strip()
        if where != "":
            sql += where + " "
        
 
         
         
        #group
        groupby = "" if self.__group == None else f" GROUP BY {self.__group}"
        groupby =  groupby.strip()
        if groupby != "":
            sql += groupby + " "
            
        orderby = "" if self.__order == None else f" ORDER BY {self.__order}"
        orderby =  orderby.strip()
        if orderby != "":
            sql += orderby + " "    
            
        #limit and offset               
        limit = f" LIMIT {self.__limit}" if self.__limit != None else ""
        limit = limit.strip()
        if limit != "":
            sql += limit + " "
        
        offset = f" OFFSET {self.__offset}" if self.__offset != None else ""
        offset = offset.strip()
        if offset != "":
            sql += offset + " "        
        
        #join
        # join = "" 
        # for jn in self.__join:
        #     join += jn
        # join = None if join == "" else join.strip()
                            
        # self.__SQL = f"SELECT {columns} FROM {self.__table} {tableAlias} {where} {limit} {offset} {groupby}"
     
        
            
        
        
        self.__SQL = sql.strip()
        
        self.__queries.append(sql)
    
    
    def __construct_insert_query(self, data):
        self.__last_method = 'insert'         
        
        data = [data] if type(data) == dict else data
        
        for row in data:
            cols = ""         
            values = ""   
            for col, value in row.items():
                keyParamValue, keyParam = self.__paramKey(col, True)               
                
                cols += f"{col}, "
                values += f"{keyParam}, "
                
        cols = cols[:-2]        
        values = values[:-2]     
                
        self.__SQL = f"INSERT INTO {self.__table} ({cols}) VALUES ({values})"    
    
    
    def __construct_update_query(self, data):
        
        self.__last_method = 'update'
        
        values = ""
        for key, value in data.items():
            keyParamValue, keyparam = self.__paramKey(key)
            self.__params[keyParamValue] = value
            
            values += f"{key} = {keyparam}, "  
                         
        values = values[:-2]   
        self.__values_update = values  
        
        
        #condicional
        where = ""
        for wheres in self.__conditional: 
            queryWhere = wheres['where']      
            where += f" {wheres['conector']} ({queryWhere})"            
        if where != '' :
            where = f" WHERE {where} "  
        where = where.replace("WHERE  AND","WHERE").replace("WHERE  OR","WHERE")
        
        self.__SQL = f"UPDATE {self.__table} SET {self.__values_update} {where}"     
    
    
    def __construct_delete_query(self):
        
        self.__last_method = 'delete'
        
        #condicional
        where = ""
        for wheres in self.__conditional: 
            queryWhere = wheres['where']      
            where += f" {wheres['conector']} ({queryWhere})"            
        if where != '' :
            where = f" WHERE {where} "  
        where = where.replace("WHERE  AND","WHERE").replace("WHERE  OR","WHERE")
        
        self.__SQL = f"DELETE FROM {self.__table} {where}"     
                  
    
    def __lenParams(self):
        return len(self.__params)
    
    
    def __paramKey(self, key, insert = False):
        pIndex = self.__lenParams()
        
        if insert == False:
            keyparam = f"{self.__table}_{key}_{pIndex}".replace('.',"_")
        else:
            keyparam = key

        return f"{keyparam}", f"%({keyparam})s"
    
    
    def upColumns(self, data):
        self.__columns = self.__columns + (data)
        
        
    def __convert_datetime_to_string(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        raise TypeError("Tipo de objeto não serializável")        


from .Querier import Querier
import json

class Modelr():
    
     
    __sofDelete__ = False   
    __table__ = None   
    __conn__ = None    
    
    
    def __init__(self) -> None:  

        self.query = Querier(conn=self.__conn__, softDelete=self.__sofDelete__).table(self.__table__)      
        
        self.__has = []      
        
        self.__queries = []     
    
    
    # get data -----------------------------------------------------------
      
    
    def first(self):        
        return self.execute('first')
    
    
    def get(self): 
        return self.execute('get')
    

    def execute(self, type:str):
        
        if type == 'get':
            data = self.query.get() 
            
        elif type == 'first':
            data = self.query.first() 
            
        elif type == 'onlysql':
            self.__queries.insert(0,{"query":self.query.toSql(),"params":self.query.toParams()})
            data = self.query.get()  
            
            print(data)
            
            if data != None:
                self.check_relationships(data, False, True) 
            return
            
            
        data = self.check_relationships(data, False)  
       
        return data 
      
      
    # get info -----------------------------------------------------------
    
    
    def toSql(self):        
        query = self.query.toSql()
        for key, value in  self.query.toParams().items():              
            query = query.replace(f"%({key})s",str(value))
                              
        return query            
    
    
    def toSqls(self):
        return self.getInfo(True)
    
    
    def toInfo(self):        
        return {"query":self.query.toSql(), "params":self.query.toParams()}
    
    
    def toInfos(self):
        return self.getInfo()
 
    
    def getInfo(self, binded = False):      
        self.execute('onlysql')
        
        if binded:
            queries = []
            for x in self.__queries:
                query = x['query']
                for key, value in x['params'].items():              
                    query = query.replace(f"%({key})s",str(value))
                
                queries.append(query)
                    
            return queries                    
                   
        else:
            return self.__queries
         
         
             
    
    # select ------------------------------------------------
        
    def select(self, columns:list):
        self.query.select(columns)
        return self
    
    
    
    # insert ---------------------------------------------------------

    def insert(self, data:dict|list[dict]):
        return self.query.insert(data)
    
    
    # update -------------------------------
    def update(self, data):
        return self.query.update(data)
    
    
    # delete ------------------------------------------------------------
    
    def delete(self):
        return self.query.delete()
        
        
    def active(self):
        return self.query.active()
    
    
    def withTrashed(self):
        self.query.withTrashed()
        return self
    
    
    
    
    # condicional-----------------------------------------------------------    
    
    def where(self, values:dict, conector:str = 'AND' ):
        self.query.where(values, conector)
        return self
        
    def whereIn(self, column:str, values:list, conector:str = 'AND', type:str = ''):
        self.query.whereIn(column, values, conector, type)
        return self
        
    def whereNotIn(self, column:str, values:list, conector:str = 'AND'):
        self.query.whereNotIn(column, values, conector)
        return self
    
    def orWhere(self, values:dict):
        self.query.orWhere(values)
        return self
    
    def whereIsNull(self, column:str, conector:str = 'AND', type:str = ''):
        self.query.whereIsNull(column, conector , type)
        return self
    
    def whereIsNotNull(self, column:str,  conector:str = 'AND'):
        self.query.whereIsNotNull(column, conector)
        return self
    
    def whereLike(self, column:str, values:list, conector:str = 'AND'):
        self.query.whereLike(column, values, conector)
        return self
    
    
    
    # pos condicional -----------------------------------
    
    def limit(self, limit:int):
        self.query.limit(limit)
        return self
    
    
    def offset(self, offset:int):
        self.query.offset(offset)
        return self
    
    
    def groupBy(self, group:str|list[str]):  
        self.query.groupBy(group)            
        return self
    
    
    def orderBy(self, order:str|list[str]):  
        self.query.orderBy(order)            
        return self
     
    

    
    
    
    # relashionship -----------------------------------------------------------
    
    
    def has(self, model:str, model_key:str, intermediate_model_key:str=None, intermediate_table:str=None, intermediate_local_key:str=None, local_key:str=None, type:str=None):
        table_local = self.__table__        
        table_model = model.__table__             
        
        for related in self.__has:
            for a in related.values():
                check = len(a)
                if check == 0:
                    key = list(related)[0]
                    related[key] = {
                        "type":type,                         
                        "local_table":table_local, 
                        "local_key":local_key, 
                        "intermediate_table":intermediate_table,
                        "model_table":table_model,
                        "model_key":model_key,     
                        "intermediate_model_key":intermediate_model_key,
                        "intermediate_table":intermediate_table,
                        "intermediate_local_key":intermediate_local_key                        
                        }
                                
    
    def hasOne(self, model, model_key, local_key): 
        self.has( 
                 model=model,
                 model_key=model_key,
                 local_key=local_key,
                 type='one'
                )
                    
       
    def hasMany(self, model, model_key, local_key):
        self.has(
                model=model,
                 model_key=model_key,
                 local_key=local_key,
                 type='many'
                 )
       
       
    def hasManyToMany(self, model, model_key, intermediate_model_key , intermediate_table, intermediate_local_key, local_key):
        self.has(
                model=model,
                model_key=model_key, 
                intermediate_model_key=intermediate_model_key,
                intermediate_table=intermediate_table,
                intermediate_local_key=intermediate_local_key,
                local_key=local_key,
                type='manytomany'
                ) 
        self.selected_related = model.__table__
        return self         
      
      
    def withPivot(self, columns:list[str]|bool = True):    
        
        for related in self.__has:
            for key in related.keys():                
                if self.selected_related == key:           
                    related[key]['pivot'] = columns
            
           
    def related(self, model:list|str):   
        
        def x(d):               
            method = getattr(self, d)       
            self.__has.append({method.__name__:{}})             
            method()  
            
        
        if type(model) == str:
            x(model)
        else:
            for i in model:
                x(i)
                
        return self
    
            
    def check_relationships(self, data, first = False, onlySql = False):   
              
        data_json = data if type(data) == list else json.loads(data)           
        data_json = data_json if type(data_json) == list else [data_json]  
        
      
        
        for dictr in self.__has:  
            for rel, value in dictr.items():        
                if value['type'] in ['many','one']:
                    data_json = self.check_relationships_one_many(rel, value, data_json, onlySql)               
            
                elif value['type'] in ['manytomany']:
                    data_json = self.check_relationships_manytomany(rel, value, data_json, onlySql)
                        
            
        if onlySql==False:   
            return data_json[0] if first else data_json
        
      
    def check_relationships_one_many(self, rel, value, data_json, onlySql):                   
                    
        all = True if value['type'] == 'many' else False                

        keys = list(set(row[value['local_key']] for row in data_json))
            
        query_related = Querier(self.__conn__).table(value['model_table']).whereIn(value['model_key'], keys)                
        self.__queries.append({"query":query_related.toSql(),"params":query_related.toParams()})            
            
        if onlySql==False:    
            data_rel = json.loads(query_related.get())                       
            if all:   
                for row in data_json:           
                    row[rel] = None                                       
                    for relrow in data_rel:        
                        if row[value['local_key']] == relrow[value['model_key']]:
                            if row[rel] == None:                                                           
                                row[rel] = []                                        
                            row[rel].append(relrow)    

                                
            else:
                for row in data_json:   
                    for relrow in data_rel: 
                        if row[value['local_key']] == relrow[value['model_key']]:
                            if rel not in row.keys():                                
                                row[rel] = relrow
                            else:      
                                row[rel] == None
                                        
        return data_json                                            
                                        
    
    def check_relationships_manytomany(self, rel, value, data_json, onlySql):
        local_keys = list(set(row[value['local_key']] for row in data_json))
   
        #acessar intermediaria
        query_intermediated = Querier(self.__conn__).table(value['intermediate_table']).whereIn(value['intermediate_local_key'], local_keys)                
        self.__queries.append({"query":query_intermediated.toSql(),"params":query_intermediated.toParams()}) 
        
        
        #resgatar ids do modelo relacionado   
        data_intermediate_json = query_intermediated.get()
        data_intermediate_json = json.loads(data_intermediate_json)           
        data_intermediate_json = data_intermediate_json if type(data_intermediate_json) == list else [data_intermediate_json]
        intemediate_model_keys = list(set(row[value['intermediate_model_key']] for row in data_intermediate_json))

    
        
        #acessar table do model   
        query_model = Querier(self.__conn__).table(value['model_table']).whereIn(value['model_key'], intemediate_model_keys)                
        self.__queries.append({"query":query_model.toSql(),"params":query_model.toParams()})    
        
        
        
        #mesclando relacionamentos
        data_rel = json.loads(query_model.get()) 
        local_counter = 0
        for local_row in data_json:          
            local_counter +=1
            
            inter_counter = 0            
            for interdata_row in data_intermediate_json:  
                inter_counter +=1
                
                if rel not in local_row:  
                    local_row[rel] = None 
                              
                model_counter = 0           
                for model_row in data_rel: 
                    model_counter +=1                   

                     
                    if local_row[value['local_key']] == interdata_row[value['intermediate_local_key']] and\
                       model_row[value['model_key']] == interdata_row[value['intermediate_model_key']]:  
                           
                        if "pivot" in value:
                            if type(value['pivot']) == list:
                                model_row['pivot'] = {col: interdata_row[col] for col in interdata_row if col in value['pivot'] and interdata_row[value['intermediate_model_key']] == model_row[value['model_key']] }
                            else:  
                                model_row['pivot'] = interdata_row 
                            
                        if local_row[rel] == None:  
                            local_row[rel] = [] 
                            
                        local_row[rel].append(model_row)       
                          
                
                    
        return data_json   


from pathlib import Path
import os,shutil
import pytest
from KassOrm.Migration import Migration  
import pdb
import time


_PATH_ = Path("_tests_results_/tests_migrations")

@pytest.fixture
def reset_migration_tests():
    dir_migration = Path(f"{_PATH_}/api/database")  
    
    Migration().drop_all_migrations(dir_migration)
    
    
    if os.path.isdir(_PATH_) == True:
        shutil.rmtree(_PATH_)
    os.makedirs(f"{_PATH_}")  
    
    
def test_create_file_migration(reset_migration_tests):       
    name_migration = "create_users" 
    dir_migration = f"{_PATH_}/api/database"
    comment = "Criação da tabela de usuários"
    table = 'users'    
    
    time.sleep(1)
    Migration().make_file_migration(name_migration, dir_migration, table, comment)
    
    assert os.path.isdir(dir_migration), f"Migration {name_migration} not created!"
    
    
# Alterar stub migration_alter.stub para funcionar     
# def test_alter_file_migration():
#     name_migration = "alter_users" 
#     dir_migration = f"{_PATH_}/api/database"
#     comment = "Criação do campo nome em users"
#     table = 'users'
    
#     time.sleep(1)    
#     Migration().make_file_migration(name_migration, dir_migration, table, comment)
    
#     assert os.path.isdir(dir_migration), f"Migration {name_migration} not created!"
    
    
def test_execute_migrations():    
    
    dir_migration = f"{_PATH_}/api/database"      
    
    result = Migration().migrate(dir_migration)
    print(result)
    
    assert result, "Migration not executed!"
    
    
# def test_rollback_one_step():
#     dir_migration = f"{_PATH_}/api/database"
        
#     result = Migration().rollback(dir_migration, 1)   
    
#     assert result, "Migration not excluded!"
    


import os, glob, importlib
from pathlib import Path
from KassOrm._helpers import getStub
from datetime import datetime as dt
from KassOrm.Querier import Querier
from KassOrm.Conn import Conn


class Migration:
    __file__ = None
    __conn__ = None
    __type__ = None
    __table__ = None
    __comment__ = None

    __rollback__ = False

    def __init__(self) -> None:
        self.__sql = ""

        self.__params = {}

        self.__columns = []

        self.__column = {}

        self.__primary_key = None
        
        self.__rollback = False
        
        self.__add_column = []


    def make_file_migration(self, name_migration, dir_migrations, table, comment: str = ""):
        """Responsável por criar os arquivos de migração"""
        name_migration = name_migration.lower()
        
        dir_migrations = Path(dir_migrations)

        if os.path.isdir(dir_migrations) == False:
            os.makedirs(dir_migrations)

        if "create" in name_migration:
            content = getStub("migration_create.stub")
        else:
            content = getStub("migration_alter.stub")

        content = content.replace(
            "%COMMENT%", f"'{comment.lower()}'" if comment != "" else "''"
        )
        content = content.replace("%TABLE%", table)

        filename = dt.now().strftime("%Y_%m_%d__%H%M%S") + "_" + name_migration + ".py"
        file = open(f"{dir_migrations}/{filename}", "w+", encoding="utf-8")
        file.writelines(content)
        file.close()


    def generate_query_create(self):
        """Gera uma query de criação de tabela"""
        self.__sql = "CREATE TABLE IF NOT EXISTS {} (".format(self.__table__)

        for col in self.__columns:
            column = ""
            for value in col.values():
                column += f"{value} "

            self.__sql += f"{column}, "

        if self.__primary_key != None:
            self.__sql += "PRIMARY KEY ({}), ".format(self.__primary_key)

        self.__sql = self.__sql[:-2]
        self.__sql += ");"
        
       
       
    def generate_query_alter(self):
        """Gera uma query de alteração de tabela"""
        
        self.__sql = "ALTER TABLE {} ".format(self.__table__)

        for col in self.__columns:
            column = ""
            for value in col.values():
                column += f"{value} "

            self.__sql += f"{column}, "


        self.__sql = self.__sql[:-2]
        


    def generate_query_rollback(self):
        """Gera uma query de alteração, drop ou exclusão de tabela"""
        
        pass 



    def generate_query(self):
        """Gerencia qual query gerar"""
        
        
        # up e down configura as colunas e o q for para gerar a query
        if self.__rollback == False:
            self.up()
        else:
            self.down()
            
            
        # gera a query de acordo com as configs de up ou down
        if self.__type__ == "create" and self.__rollback == False:
           self.generate_query_create()

        elif self.__type__ == "alter": 
            self.generate_query_alter()


    def setRollback(self):
        self.__rollback = True
        return self


    def execute(self):
        """Principal método de execução da migration, gera o sql do arquivo migration e o executa"""
 
        try:
            self.generate_query()  
            
            result = Conn().getQuery(self.__sql, self.__params).execute_create()
            return result
        except Exception as err:
            print("execute: " + str(err))
            return False


    def execute_migrate(self, module_name:str):
        """Executa o metodo execute da classe migrate do arquivo migration""" 
        module = importlib.import_module(module_name)
        return module.migrate(), module.migrate().execute()
    
    
    def execute_rollback(self, module_name:str):
        module = importlib.import_module(module_name)
        return module.migrate(), module.migrate().setRollback().execute()
    
    
    def catch_module_and_migrations(self, dir_migrations):
        """Retorna o nome do diretório como módulo e todas as migrações do diretorio"""
        migrations = glob.glob(os.path.join(dir_migrations, "*.py"))
        dir_module = str(dir_migrations).replace("\\", ".")
        return dir_module, migrations


    def has_migration_executed(self, migration:str):
        """Verifica se a migração foi executada, salva na tabela _migrations_"""
        try:
            return Querier().table("_migrations_").where({"migration":migration}).first()
        except Exception as err:
            print(str(err))
            return False
     
     
    def save_migration_executed(self, migration: str, description: str = None):
        """Salva registro da migração executada na tabela _migrations_"""
        try:
            # sql = """INSERT INTO _migrations_ (date, migration, description) VALUES (NOW(), %(migration)s, %(description)s)"""
            # Conn().getQuery(
            #     sql, {"migration": migration, "description": description}
            # ).execute_insert()

            now = dt.now() 
            Querier().table('_migrations_').insert({"date":now,"migration":migration,"description":description})
            return True
        except Exception as err:
            print("save_migration_executed:"  +str(err))
            return False
    
    
    def delete_migration_executed(self, id):
        
        Querier().table('_migrations_').where({"id":id}).delete()
        return True


    def execute_all_migrations(self, dir_migrations: str):
        """Executa todas as migrações do diretorio informado"""

        dir_module, migrations = self.catch_module_and_migrations(dir_migrations)
            
        step = 1
        for migration in migrations:
            file = os.path.basename(migration).replace(".py", "")  
            
            
            if self.has_migration_executed(file) == 'null':       
                module, result = self.execute_migrate(f"{dir_module}.{file}")

                if result == True:
                    self.save_migration_executed(file, module.__comment__)
                    print(f"{file} {self.color('green')}[OK]{self.color()}")
                else:
                    print(f"{file} {self.color('red')}[Fail]{self.color()}")
                    
                    self.rollback(dir_migrations, step)
                    return False

            else:
                print(f"{file} {self.color('yellow')}[ALREADY EXISTS]{self.color()}")

            step +=1

        
        return True


    def drop_all_migrations(self, dir_migrations: str):
        migrations = glob.glob(os.path.join(dir_migrations, "*.py"))

        self.drop_table_migrations()

        dir_module = str(dir_migrations).replace("\\", ".")

        for migration in migrations:
            file = os.path.basename(migration).replace(".py", "")
            
            module, result = self.execute_rollback(f"{dir_module}.{file}")            

            print(f"{file} {self.color('yellow')} [DROPPED]" + self.color())
            
        return True



    def make_file_model(self, dir_models, table, comment: str = ""):
        """Responsável por criar os arquivos de migração"""
        # table = table.
        
        dir_models = Path(dir_models)
        # modelname = table.replace("_"," ").capitalize().replace(" ","")
        
        parts = table.split('_')
        modelname = parts[0] + ''.join(x.title() for x in parts[1:])

        if os.path.isdir(dir_models) == False:
            os.makedirs(dir_models)

        content = getStub("model.stub")


        content = content.replace("%COMMENT%", f"'{comment.lower()}'" if comment != "" else "''")
        content = content.replace("%TABLE%", table)
        content = content.replace("%MODELNAME%", modelname)
        

        filename = modelname + ".py"
        file = open(f"{dir_models}/{filename}", "w+", encoding="utf-8")
        file.writelines(content)
        file.close()












    def up(self):
        return self

    def down(self):
        return self

    #  TIPOSS ------------------

    def add(self):
        self.__columns.append(self.__column)
        self.__column = {}
        return

    def id(self):
        self.__column["name"] = "id"
        self.__column["type"] = "BIGINT"
        self.__column["unsigned"] = "UNSIGNED"
        self.__column["isNull"] = "NOT NULL"
        self.__column["autoIncrement"] = "AUTO_INCREMENT"
        self.__column["primary_key"] = "PRIMARY KEY"
        
        

        return self

    def string(self, name, qnt: int = 255):
        self.__column["name"] = name
        self.__column["type"] = f"VARCHAR({qnt})"
        self.__column["isNull"] = "NOT NULL"

        return self

    def bigInteger(self, name):
        self.__column["name"] = name
        self.__column["type"] = f"BIGINT"
        self.__column["isNull"] = "NOT NULL"

        return self

    def bigIntegerUnisigned(self, name):
        self.__column["name"] = name
        self.__column["type"] = f"BIGINT UNSIGNED"
        self.__column["isNull"] = "NOT NULL"

        return self

    def text(self, name):
        self.__column["name"] = name
        self.__column["type"] = f"TEXT"
        self.__column["isNull"] = "NOT NULL"

        return self

    def enum(self, name: str, values: list):
        str_values = ", ".join(map(lambda x: f'"{x}"', values))

        self.__column["name"] = name
        self.__column["type"] = f"ENUM({str_values})"
        self.__column["isNull"] = "NOT NULL"

        return self

    def integer(self, name: str, qnt: int = None):
        self.__column["name"] = name

        if qnt != None:
            self.__column["type"] = f"INT({qnt}) ZEROFILL UNSIGNED"
        else:
            self.__column["type"] = f"INT"
        self.__column["isNull"] = "NOT NULL"

        return self

    def datetime(self, name: str):
        self.__column["name"] = name
        self.__column["type"] = "DATETIME"
        self.__column["isNull"] = "NOT NULL"

        return self

    def nullable(self):
        self.__column["isNull"] = "NULL"

        return self

    def unsigned(self):
        self.__column["unsigned"] = "UNSIGNED"

        return self

    def comment(self, comment: str):
        self.__column["comment"] = f"COMMENT '{comment}' "
        return self

    def unique(self, columns: list = None, name=None):
        if columns != None or name != None:
            uniq = ""
            for col in columns:
                uniq += f"{col}, "

            uniq = uniq[:-2]

            self.__column["unique_key"] = f" UNIQUE KEY {name} ()"
            return self
        else:
            self.__column["unique"] = f"UNIQUE"
            return self

    def current_timestamp(self):
        self.__column["current_timestamp"] = "DEFAULT CURRENT_TIMESTAMP"

        return self

    def update_timestamp(self):
        self.__column["on_update_timestamp"] = "ON UPDATE CURRENT_TIMESTAMP"

        return self

    # TOOLS-----------

    def addColumn(self):
        self.__column["add"] = "ADD COLUMN "

        return self

    def dropColumn(self, name):
        self.__column["drop"] = "DROP COLUMN {} ".format(name)
        return self

    def after(self, column: string):
        self.__column["after"] = f"AFTER {column}"
        return self

    def first(self, column: string):
        self.__column["first"] = f"FIRST {column}"
        return self

    def dropTableIfExists(self):
        self.__sql = "DROP TABLE IF EXISTS {}".format(self.__table__)

    def color(self, code=""):
        """Cores para usar no terminal"""

        if code == "green":
            return "\033[0;32m"

        elif code == "yellow":
            return "\033[0;33m"

        elif code == "red":
            return "\033[0;31m"

        else:
            return "\033[m"

    def create_table_migrations(self):
        try:
            sql = """
                CREATE table if not exists _migrations_ (
                id BIGINT unsigned not null auto_increment,
                date DATETIME not null,
                migration VARCHAR(255) not null unique,
                description VARCHAR(255) null,
                PRIMARY KEY (`id`)
                );
            """
            Conn().getQuery(sql, {}).execute_create()
            return True
        except Exception as err:
            print("create_table_migrations: "+str(err))
            return False

    def drop_table_migrations(self):
        try:
            sql = "DROP TABLE IF EXISTS _migrations_"
            Conn(None).getQuery(sql, None).execute_create()
            return True
        except Exception as err:
            print("drop_table_migrationsstr: "+(err))
            return False

 

  

    def migrate(self, dir_migrations):
        
        dir_migrations = Path(dir_migrations)
        hasCreated = self.create_table_migrations()

        if hasCreated==False:
            return False
        
        result = self.execute_all_migrations(dir_migrations)

        return  result




    def rollback(self, dir_migrations, steps):
       
       dir_migrations = Path(dir_migrations) 
        
       migrations = Querier().table('_migrations_').orderBy("id DESC").limit(steps).get()
       
       dir_module = str(dir_migrations).replace("\\", ".")
        
       for migration in migrations:
            file = migration['migration']            

            module, result = self.execute_rollback(f"{dir_module}.{file}")
            
            if result == False:
                print(result)
                return False
            
            self.delete_migration_executed(migration['id'] )


       
       return True


from setuptools import setup, find_packages

setup(
    name='KassOrm',
    version='1.0.1',
    packages=find_packages(),
    install_requires=[
        "mysql-connector-python",'python-dotenv'
        ],
    tests_require=[
        'pytest',
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points={    },
    author='Kássio Douglas',
    author_email='kass.doug@gmail.com',
    description='Gerenciar banco de dados',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/kassiodouglas/KassOrm',
    license='MIT',
)
